using System;
using System.Collections.Generic;
using System.Net;
using System.Net.Sockets;
using System.IO;
using System.Text;
using System.Threading;
using System.Windows.Forms;
using HBDotNet;
using System.Data;
using System.Data.OleDb;
using Microsoft.Win32;

namespace TitleManagerHB
{
    [HBDevice("TheaterTek Title Manager")]
    public class Class1 : HBDeviceBase
    {
        private OleDbConnection conn;
        private TitleManagerHB.TheaterTekDataSet ttDS;
        private OleDbDataAdapter ttDA;
        private DataView ttDV;
        private Thread _threadDB;

        public override bool OnInit(bool enabled)
        {
            if (enabled) OnEnable();
            return base.OnInit(enabled);
        }

        public override void OnShutdown()
        {
            base.OnShutdown();
        }

        public override bool OnEnable()
        {
            //System.Diagnostics.Trace.WriteLine("OnEnable");
            // Initialize the connection.  Since OnInit also calls OnEnable, I only need to have this call here.
            if (!ConnectToDatabase())
                return false;
            else
                return base.OnEnable();
        }

        public override bool OnDisable()
        {
            CloseDatabase();
            return base.OnDisable();
        }

        public override void OnAboutBox()
        {
            MessageBox.Show("This plugin implements a database-driven title manager for the TheaterTek Control plugin.");
        }

        protected override bool OnPropertyChangeRequest(HBDotNet.HBProperty hbProperty, string newValue)
        {
            if (hbProperty.Name == "TheaterTek Title List Filter")
                PopulateTitleList(newValue, "");

            if (hbProperty.Name == "TheaterTek Title List Sort By")
                PopulateTitleList("", newValue);

            if (hbProperty.Name == "TheaterTek Title Database ID")
            {
                string titlePath = "";
                ttDV.RowFilter = "ID = '" + newValue + "'";

                foreach (DataRowView drv in ttDV)
                {
                    titlePath = Convert.ToString(drv["Path"]);
                }

                if (Directory.Exists(titlePath + "\\VIDEO_TS")) titlePath += "\\VIDEO_TS";

                Properties.GetPropertyByName("TheaterTek Title Path").CurrentValue = titlePath;
            }

            if (hbProperty.Name == "Title Manager Command" && newValue == "Play Selected Title")
            {
                SetTheaterTekPath(Properties.GetPropertyByName("TheaterTek Title Path").CurrentValue);
            }

            if (hbProperty.Name == "Title Manager Command" && newValue == "Update Database")
            {
                _threadDB = new Thread(UpdateDatabase);
                _threadDB.Start();
            }

            return base.OnPropertyChangeRequest(hbProperty, newValue);
        }

        public void SetTheaterTekPath(string newPath)
        {
            IntPtr ttHandle;
            bool ttEnabled;
            string ttDeviceName = "";

            StringBuilder sb = new StringBuilder();
            sb.Append("TheaterTek Title Path");
            
            HBModule.GetDeviceInfoFromDescription("TheaterTek", out ttHandle, out ttEnabled, out ttDeviceName);
            IntPtr ttIDHandle = HBModule.GetDevicePropertyHandleFromName(ttHandle, "TheaterTek Title Path", sb);
            IntPtr hPropertyValue = HBModule.CreatePropertyValue(ttIDHandle, newPath);
            HBModule.ChangeRequest(hPropertyValue);
        }

        public override bool OnRegisterProperties()
        {
            //System.Diagnostics.Trace.WriteLine("OnRegisterProperties");
            Properties.Add(new HBProperty("TheaterTek Title", "TheaterTek Title", true, "", eIOType.ioInputOnly, true, true, true, false, ePropertyType.ptAlpha, ""));
            Properties.Add(new HBProperty("TheaterTek Title Path", "TheaterTek Title Path", false, "", eIOType.ioInputAndOutput, true, true, true, false, ePropertyType.ptAlpha, ""));
            Properties.Add(new HBProperty("TheaterTek Title Database ID", "TheaterTek Title Database ID", false, "", eIOType.ioInputAndOutput, true, true, true, false, ePropertyType.ptNumeric, ""));

            Properties.Add(new HBProperty("TheaterTek Title List", "TheaterTek Title List", false, "", eIOType.ioInputAndOutput, true, true, true, false, ePropertyType.ptAlphaList, ""));
            Properties.Add(new HBProperty("TheaterTek Title List Filter", "TheaterTek Title List Filter", false, "", eIOType.ioInputAndOutput, true, true, true, false, ePropertyType.ptAlpha, ""));
            Properties.Add(new HBProperty("TheaterTek Title List Sort By", "TheaterTek Title List Sort By", false, "Alphabetical", eIOType.ioInputAndOutput, true, true, true, false, ePropertyType.ptAlpha, ""));

            Properties.Add(new HBProperty("Title Manager Command", "Title Manager Command", false, "", eIOType.ioInputAndOutput, true, true, true, false, ePropertyType.ptAlpha, ""));
            Properties.Add(new HBProperty("Title Manager Database Status", "Title Manager Database Status", false, "", eIOType.ioInputOnly, true, true, true, false, ePropertyType.ptAlpha, ""));
            Properties.Add(new HBProperty("Title Manager Directory List", "Title Manager Directory List", false, "", eIOType.ioInputAndOutput, true, true, true, false, ePropertyType.ptAlphaList, ""));
            Properties.Add(new HBProperty("Title Manager Database Title Count", "Title Manager Database Title Count", false, "", eIOType.ioInputOnly, true, true, true, false, ePropertyType.ptNumeric, ""));

            HBModule.CreateDevicePropertyValue("TheaterTek Title List Filter", "Movies");
            HBModule.CreateDevicePropertyValue("TheaterTek Title List Filter", "TV Shows");
            HBModule.CreateDevicePropertyValue("TheaterTek Title List Filter", "Music");
            //HBModule.CreateDevicePropertyValue("TheaterTek Title List Filter", "ReplayTV Show");
            HBModule.CreateDevicePropertyValue("TheaterTek Title List Filter", "Extras");
            HBModule.CreateDevicePropertyValue("TheaterTek Title List Filter", "All");

            HBModule.CreateDevicePropertyValue("TheaterTek Title List Sort By", "Alphabetical");
            HBModule.CreateDevicePropertyValue("TheaterTek Title List Sort By", "Oldest First");
            HBModule.CreateDevicePropertyValue("TheaterTek Title List Sort By", "Newest First");

            HBModule.CreateDevicePropertyValue("Title Manager Command", "Play Selected Title");
            HBModule.CreateDevicePropertyValue("Title Manager Command", "Update Database");
            HBModule.CreateDevicePropertyValue("Title Manager Command", "GUI Filler");

            return base.OnRegisterProperties();
        }

        public bool ConnectToDatabase()
        {
            string hbPath = "";
            string dbPath = "";
            RegistryKey masterKey = Registry.LocalMachine.CreateSubKey("Software\\CeBotics\\HouseBot\\General Settings");

            if (masterKey != null)
            {
                hbPath = masterKey.GetValue("Root Application Directory").ToString();
            }
            masterKey.Close();

            if (hbPath != "") dbPath = hbPath + "Plugins\\Devices\\TheaterTek.mdb";
            if (dbPath == "") dbPath = "C:\\TheaterTek.mdb";

            string source = "Provider=Microsoft.Jet.OLEDB.4.0;Data Source=" + dbPath + ";User Id=admin;Password=;";
            conn = new OleDbConnection(source);
            conn.Open();

            ttDS = new TheaterTekDataSet();
            ttDA = new OleDbDataAdapter("SELECT * FROM Titles", conn);
            ttDA.Fill(ttDS, "Titles");
            ttDV = new DataView(ttDS.Titles);
            PopulateTitleList("Movies", "Title ASC");
            Properties.GetPropertyByName("TheaterTek Title List Filter").CurrentValue = "Movie";

            OleDbCommand insCmd = new OleDbCommand("INSERT INTO Titles (ID, Title, Path, Category, DateRipped) VALUES (@ID, @Title, @Path, @Category, @DateRipped)", conn);
            insCmd.Parameters.Add("@ID", OleDbType.Integer, 4, "ID");
            insCmd.Parameters.Add("@Title", OleDbType.VarWChar, 255, "Title");
            insCmd.Parameters.Add("@Path", OleDbType.VarWChar, 255, "Path");
            insCmd.Parameters.Add("@Category", OleDbType.VarWChar, 255, "Category");
            insCmd.Parameters.Add("@DateRipped", OleDbType.DBDate, 4, "DateRipped");
            ttDA.InsertCommand = insCmd;

            OleDbCommand delCmd = new OleDbCommand("DELETE FROM Titles WHERE ID=@ID", conn);
            delCmd.Parameters.Add("@ID", OleDbType.Integer, 4, "ID");
            ttDA.DeleteCommand = delCmd;

//            DataSet dsTemp = new DataSet();
//            ttDA.SelectCommand = "SELECT DISTINCT Category FROM Titles";
//            ttDA.Fill(dsTemp);
//            foreach (DataRow dr in dsTemp)
//            {
////                HBModule.CreateDevicePropertyValue("
//            }
            return true;
        }

        public void CloseDatabase()
        {
            ttDV.Dispose();
            ttDA.Dispose();
            ttDS.Dispose();
            conn.Close();
        }

        private void UpdateDatabase()
        {
            Properties.GetPropertyByName("Title Manager Database Status").CurrentValue = "Updating database (started " + System.DateTime.Now.ToString() + ")";

            // First we'll delete all existing rows in the database
            foreach (TheaterTekDataSet.TitlesRow tr in ttDS.Titles)
            {
                tr.Delete();
            }
            ttDA.Update(ttDS.Titles);

            // Now we'll populate the database getting the list of directories to scan from a property
            string[] directoryList;
            string category = "";
            directoryList = Properties.GetPropertyByName("Title Manager Directory List").CurrentValue.Split((char)10);
            foreach (string dirName in directoryList)
            {
                if (Directory.Exists(dirName))
                {
                    category = dirName.Substring(dirName.LastIndexOf("\\") + 1);

                    AddTitlesToDatabase(dirName, category);
                }
            }
            // \\vader\media\dvds\Movies
            // \\vader\media\dvds\TV Shows
            // \\vader\media\dvds\Music
            // \\htpc\media\TV Shows
            // \\htpc\media\Extras

            //AddReplayTVShowsToDatabase(@"\\vader\media\ReplayTV Shows\Local_Guide", "ReplayTV Show");
            ttDA.Update(ttDS.Titles);

            Properties.GetPropertyByName("Title Manager Database Title Count").CurrentValue = ttDS.Titles.Rows.Count.ToString();
            Properties.GetPropertyByName("Title Manager Database Status").CurrentValue = "Database updated (finished " + System.DateTime.Now.ToString() + ")";
            return;
        }

        private void PopulateTitleList(string filterValue, string sortValue)
        {
            Properties.GetPropertyByName("TheaterTek Title List").CurrentValue = "** working **";

            if (filterValue != string.Empty)
                if (filterValue != "All")
                    ttDV.RowFilter = "Category LIKE '" + filterValue + "'";
                else
                    ttDV.RowFilter = "Title LIKE '%'";
            else
                ttDV.RowFilter = "Category LIKE '" + Properties.GetPropertyByName("TheaterTek Title List Filter").CurrentValue + "'";

            switch (sortValue)
            {
                case "Alphabetical":
                    ttDV.Sort = "Title ASC";
                    break;

                case "Oldest First":
                    ttDV.Sort = "DateRipped ASC";
                    break;

                case "Newest First":
                    ttDV.Sort = "DateRipped DESC";
                    break;

                default:
                    ttDV.Sort = "Title ASC";
                    Properties.GetPropertyByName("TheaterTek Title List Sort By").CurrentValue = "Alphabetical";
                    break;
            }

            StringBuilder sb = new StringBuilder();
            foreach (DataRowView drv in ttDV)
            {
                sb.Append(drv["ID"] + "\t" + drv["Title"] + "\n");
            }
            Properties.GetPropertyByName("TheaterTek Title List").CurrentValue = sb.ToString();
        }

        private void AddTitlesToDatabase(string directoryName, string category)
        {
            string[] titleDirectories;
            titleDirectories = Directory.GetDirectories(directoryName);
            foreach (string movie in titleDirectories)
            {
                string movieTitle = movie.Substring(movie.LastIndexOf("\\") + 1);
                TheaterTekDataSet.TitlesRow tr;
                tr = ttDS.Titles.NewTitlesRow();
                tr.Title = movieTitle;
                tr.Path = movie;
                tr.Category = category;
                tr.DateRipped = Directory.GetCreationTime(movie);
                ttDS.Titles.Rows.Add(tr);
            }
        }

        private void AddReplayTVShowsToDatabase(string directoryName, string category)
        {
            string[] titleNames;
            titleNames = Directory.GetFiles(directoryName, "*.mpg");

            foreach (string movie in titleNames)
            {
                string movieTitle = movie.Substring(movie.LastIndexOf("\\") + 1);
                movieTitle = movieTitle.Substring(0, movieTitle.Length - 4);    // remove the .mpg
                TheaterTekDataSet.TitlesRow tr;
                tr = ttDS.Titles.NewTitlesRow();
                tr.Title = movieTitle;
                tr.Path = movie;
                tr.Category = category;
                tr.DateRipped = File.GetCreationTime(movie);
                ttDS.Titles.Rows.Add(tr);
            }
        }

    }
}
