﻿using System;
using System.Collections.Generic;
using System.Net;
using System.Net.Sockets;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
using System.Windows.Forms;
using HBDotNet;

namespace TheaterTek
{
    [HBDevice("TheaterTek")]
    public class Class1 : HBDeviceBase
    {
        private Thread _thread;

        private string responseString = "";
        private string propertyValue = "";
        private string ttAddress = "";
        private int ttPort = 0;
        private int ttAKPort = 0;

        byte[] m_DataBuffer = new byte[100];
        IAsyncResult m_asynResult;
        public AsyncCallback pfnCallBack;
        public Socket m_socClient;

        private bool InitConnection()
        {
            ttAddress = Properties.GetPropertyByName("TheaterTek IP Address").CurrentValue;
            ttPort = Convert.ToInt16(Properties.GetPropertyByName("TheaterTek IP Port").CurrentValue);
            ttAKPort = Convert.ToInt16(Properties.GetPropertyByName("TheaterTek Autokiller IP Port").CurrentValue);
            if (ttAddress == "") ttAddress = "192.168.1.105";
            if (ttPort == 0) ttPort = 2663;
            if (ttAKPort == 0) ttAKPort = 2664;

            try
            {
                OpenSocket(ttAddress, ttPort);
            }
            catch (Exception)
            {
                HBModule.TraceDeviceMessage(_deviceHandle, TRC_TYPE.ttError, "Error opening socket to TheaterTek host PC!");
                return false;
            }

            // TODO: here is where I will call the code to start TT if needed
            // The sleep commands below are needed to prevent some commands from being lost (like what used to
            // happen on the Pronto remote in macros)
            ControlTT(TheaterTekMessages.IP_DVDCHAPTER);
            Thread.Sleep(200);
            ControlTT(TheaterTekMessages.IP_MEDIASTATE);
            Thread.Sleep(200);
            ControlTT(TheaterTekMessages.IP_MEDIATIME);
            Thread.Sleep(200);
            ControlTT(TheaterTekMessages.IP_CURFILE);
            Thread.Sleep(200);
            ControlTT(TheaterTekMessages.IP_DVDCHAPTERCOUNT);

            return true;
        }

        public override bool OnInit(bool enabled)
        {
            if (enabled) OnEnable();
            return base.OnInit(enabled);
        }

        public override void OnShutdown()
        {
            CloseSocket();
            if (_thread != null) _thread.Abort();
            base.OnShutdown();
        }

        public override bool OnEnable()
        {
            // Initialize the connection.  Since OnInit also calls OnEnable, I only need to have this call here.
            if (!InitConnection())
                return false;
            else
                return base.OnEnable();
            // TODO: should I put the socket IO in a separate thread?  should I put the polling in a separate thread?
            //       if so I would just move the InitConnection method into the thread.
            //_thread = new Thread(ThreadFunc);
            //_thread.Start();
        }

        public override bool OnDisable()
        {
            CloseSocket();
            if (_thread != null) _thread.Abort();
            _thread = null;
            return base.OnDisable();
        }

        public override void OnAboutBox()
        {
            MessageBox.Show("This plugin implements a two-way network interface to a PC running TheaterTek software.");
        }

        public void ThreadFunc()
        {
            while (true)
            {
                Thread.Sleep(1000);
                DateTime D = DateTime.Now;
                //Properties.GetPropertyByName("Current Time").CurrentValue = D.ToString();
                //if ((D.Second % 5) == 0)
                //{
                //    if (Properties.GetPropertyByName("Tick Every 5 Seconds").CurrentValue == "Yes")
                //        System.Diagnostics.Debug.WriteLine("BEEP");
                //}
            }
        }

        protected override bool OnPropertyChangeRequest(HBDotNet.HBProperty hbProperty, string newValue)
        {
            if (hbProperty.Name == "TheaterTek Command" | hbProperty.Name == "A/V Status")
                ControlTT(newValue);

            if (hbProperty.Name == "TheaterTek Chapter")
            {
                ControlTT(TheaterTekMessages.IP_DVDPLAYCHAPTER + " " + newValue);
            }

            if (hbProperty.Name == "TheaterTek Title Path")
            {
                if (Directory.Exists(newValue))
                {
                    // if the new value doesn't contain \VIDEO_TS but that directory exists, then add it to the path
                    if (newValue.ToLower().IndexOf("\\video_ts") == -1 && Directory.Exists(newValue + "\\VIDEO_TS")) newValue += "\\VIDEO_TS";
                    ControlTT(TheaterTekMessages.IP_PLAYFILE + " " + newValue);
                }
                else
                {
                    HBModule.TraceDeviceMessage(_deviceHandle, TRC_TYPE.ttError, "Non-existent TheaterTek title path specified.");
                }
            }

            return base.OnPropertyChangeRequest(hbProperty, newValue);
        }

        public override bool OnRegisterProperties()
        {
            Properties.Add(new HBProperty("TheaterTek Title", "TheaterTek Title", true, "", eIOType.ioInputOnly, true, true, true, false, ePropertyType.ptAlpha, ""));
            Properties.Add(new HBProperty("TheaterTek Time", "TheaterTek Time", false, "", eIOType.ioInputOnly, true, true, true, false, ePropertyType.ptAlpha, ""));
            Properties.Add(new HBProperty("A/V Status", "Status", false, "", eIOType.ioInputAndOutput, true, true, true, false, ePropertyType.ptAlpha, ""));
            Properties.Add(new HBProperty("TheaterTek Chapter", "TheaterTek Chapter", false, "", eIOType.ioInputAndOutput, true, true, true, false, ePropertyType.ptNumeric, ""));
            Properties.Add(new HBProperty("TheaterTek Chapter Count", "TheaterTek Chapter Count", false, "", eIOType.ioInputOnly, true, true, true, false, ePropertyType.ptNumeric, ""));
            Properties.Add(new HBProperty("TheaterTek Command", "TheaterTek Command", false, "", eIOType.ioInputAndOutput, true, true, true, false, ePropertyType.ptAlpha, ""));
            Properties.Add(new HBProperty("TheaterTek Title Path", "TheaterTek Title Path", false, "", eIOType.ioInputAndOutput, true, true, true, false, ePropertyType.ptAlpha, ""));

            Properties.Add(new HBProperty("TheaterTek IP Address", "TheaterTek IP Address", false, "", eIOType.ioInputAndOutput, true, true, true, false, ePropertyType.ptAlpha, "1:Enter the IP Address used for the TheaterTek server"));
            Properties.Add(new HBProperty("TheaterTek IP Port", "TheaterTek IP Port", false, "2663", eIOType.ioInputAndOutput, true, true, true, false, ePropertyType.ptNumeric, "2:Enter the IP Port used by the TheaterTek server (typically 2663)"));
            Properties.Add(new HBProperty("TheaterTek Autokiller IP Port", "TheaterTek Autokiller IP Port", false, "2664", eIOType.ioInputAndOutput, true, true, true, false, ePropertyType.ptNumeric, "3:Enter the Autokiller IP Port used by the TheaterTek server (typically 2664)"));

            Properties.Add(new HBProperty("TheaterTek Running?", "TheaterTek Running?", false, "Unknown", eIOType.ioInputOnly, true, true, true, false, ePropertyType.ptAlpha, ""));
            Properties.Add(new HBProperty("TheaterTek PC Reachable?", "TheaterTek PC Reachable?", false, "Unknown", eIOType.ioInputOnly, true, true, true, false, ePropertyType.ptAlpha, ""));

            HBModule.CreateDevicePropertyValue("TheaterTek Command", "Next");
            HBModule.CreateDevicePropertyValue("TheaterTek Command", "Previous");
            HBModule.CreateDevicePropertyValue("TheaterTek Command", "DVD Menu");
            HBModule.CreateDevicePropertyValue("TheaterTek Command", "Up");
            HBModule.CreateDevicePropertyValue("TheaterTek Command", "Down");
            HBModule.CreateDevicePropertyValue("TheaterTek Command", "Left");
            HBModule.CreateDevicePropertyValue("TheaterTek Command", "Right");
            HBModule.CreateDevicePropertyValue("TheaterTek Command", "Enter");
            HBModule.CreateDevicePropertyValue("TheaterTek Command", "Replay");
            HBModule.CreateDevicePropertyValue("TheaterTek Command", "Jump");
            HBModule.CreateDevicePropertyValue("TheaterTek Command", "Set Movie Start");
            HBModule.CreateDevicePropertyValue("TheaterTek Command", "Chapter List");
            HBModule.CreateDevicePropertyValue("TheaterTek Command", "Audio Menu");
            HBModule.CreateDevicePropertyValue("TheaterTek Command", "Next Audio");

            HBModule.CreateDevicePropertyValue("TheaterTek Command", "REW 1X");
            HBModule.CreateDevicePropertyValue("TheaterTek Command", "REW 2X");
            HBModule.CreateDevicePropertyValue("TheaterTek Command", "REW 4X");
            HBModule.CreateDevicePropertyValue("TheaterTek Command", "REW 8X");
            HBModule.CreateDevicePropertyValue("TheaterTek Command", "REW 20X");

            HBModule.CreateDevicePropertyValue("TheaterTek Command", "FF 2X");
            HBModule.CreateDevicePropertyValue("TheaterTek Command", "FF 4X");
            HBModule.CreateDevicePropertyValue("TheaterTek Command", "FF 8X");
            HBModule.CreateDevicePropertyValue("TheaterTek Command", "FF 20X");

            HBModule.CreateDevicePropertyValue("TheaterTek Command", "Slow");
            HBModule.CreateDevicePropertyValue("TheaterTek Command", "Resume");
            HBModule.CreateDevicePropertyValue("TheaterTek Command", "Info");

            HBModule.CreateDevicePropertyValue("TheaterTek Command", "Start TheaterTek");

            return base.OnRegisterProperties();
        }

        public void ControlTT(string commandToSend)
        {
            switch (commandToSend)
            {
                case "Playing":
                    SendCommandToTT(TheaterTekMessages.ID_PLAY);
                    break;

                case "Paused":
                    SendCommandToTT(TheaterTekMessages.ID_PAUSE);
                    break;

                case "Stopped":
                    SendCommandToTT(TheaterTekMessages.ID_STOP);
                    break;

                case "Fast Forwarding":
                    UnpauseIfNeeded();
                    SendCommandToTT(TheaterTekMessages.ID_FF_5X);
                    break;

                case "Fast Rewinding":
                    UnpauseIfNeeded();
                    SendCommandToTT(TheaterTekMessages.ID_RW_5X);
                    break;

                case "Next":
                    SendCommandToTT(TheaterTekMessages.ID_NEXT);
                    break;

                case "Previous":
                    SendCommandToTT(TheaterTekMessages.ID_PREVIOUS);
                    break;

                case "DVD Menu":
                    SendCommandToTT(TheaterTekMessages.ID_ROOT_MENU);
                    break;

                case "Up":
                    SendCommandToTT(TheaterTekMessages.ID_K_UP);
                    break;

                case "Down":
                    SendCommandToTT(TheaterTekMessages.ID_K_DOWN);
                    break;

                case "Left":
                    SendCommandToTT(TheaterTekMessages.ID_K_LEFT);
                    break;

                case "Right":
                    SendCommandToTT(TheaterTekMessages.ID_K_RIGHT);
                    break;

                case "Enter":
                    SendCommandToTT(TheaterTekMessages.ID_K_RETURN);
                    break;

                case "Escape":
                    SendCommandToTT(TheaterTekMessages.ID_PREVIOUS);
                    break;

                case "Replay":
                    SendCommandToTT(TheaterTekMessages.ID_REPLAY);
                    break;

                case "Jump":
                    SendCommandToTT(TheaterTekMessages.ID_JUMP);
                    break;

                case "Set Movie Start":
                    SendCommandToTT(TheaterTekMessages.ID_SETSTARTPOINT);
                    break;

                case "Chapter List":
                    SendCommandToTT(TheaterTekMessages.ID_MENU_CHAPTER);
                    break;

                case "Audio Menu":
                    SendCommandToTT(TheaterTekMessages.ID_MENU_AUDIO);
                    break;

                case "Next Audio":
                    SendCommandToTT(TheaterTekMessages.ID_NEXTAUDIO);
                    break;

                case "REW 1X":
                    UnpauseIfNeeded();
                    SendCommandToTT(TheaterTekMessages.ID_RW_1X);
                    break;

                case "REW 2X":
                    UnpauseIfNeeded();
                    SendCommandToTT(TheaterTekMessages.ID_RW_2X);
                    break;

                case "REW 4X":
                    UnpauseIfNeeded();
                    SendCommandToTT(TheaterTekMessages.ID_RW_5X);
                    break;

                case "REW 8X":
                    UnpauseIfNeeded();
                    SendCommandToTT(TheaterTekMessages.ID_RW_10X);
                    break;

                case "REW 20X":
                    UnpauseIfNeeded();
                    SendCommandToTT(TheaterTekMessages.ID_RW_20X);
                    break;

                case "FF 2X":
                    UnpauseIfNeeded();
                    SendCommandToTT(TheaterTekMessages.ID_FF_2X);
                    break;

                case "FF 4X":
                    UnpauseIfNeeded();
                    SendCommandToTT(TheaterTekMessages.ID_FF_5X);
                    break;

                case "FF 8X":
                    UnpauseIfNeeded();
                    SendCommandToTT(TheaterTekMessages.ID_FF_10X);
                    break;

                case "FF 20X":
                    UnpauseIfNeeded();
                    SendCommandToTT(TheaterTekMessages.ID_FF_20X);
                    break;

                case "Slow":
                    UnpauseIfNeeded();
                    SendCommandToTT(TheaterTekMessages.ID_SLOW);
                    break;

                case "Resume":
                    SendCommandToTT(TheaterTekMessages.ID_BK_RESUME);
                    break;

                case "Info":
                    SendCommandToTT(TheaterTekMessages.ID_INFO);
                    break;

                default:
                    SendCommandToTT(commandToSend);
                    break;
            }
            // for any of the REW or FF commands, change the status property to the appropriate value
            if (commandToSend.Length > 2 && commandToSend.Substring(0, 3) == "REW")
                Properties.GetPropertyByName("A/V Status").CurrentValue = "Fast Rewinding";

            if (commandToSend.Length > 1 && commandToSend.Substring(0, 2) == "FF")
                Properties.GetPropertyByName("A/V Status").CurrentValue = "Fast Forwarding";
        }

        private void UnpauseIfNeeded()
        {
            if (Properties.GetPropertyByName("A/V Status").CurrentValue == "Paused")
            {
                // TT won't FF or REW if currently paused.  In that case, we'll first play, then sleep
                // (to allow time for the "Play" response to come back) and then FF/REW.
                SendCommandToTT(TheaterTekMessages.ID_PLAY);
                Thread.Sleep(500);
            }
        }

        private void parseResponse(string responseString)
        {
            //System.Diagnostics.Trace.WriteLine("Response is " + responseString);

            string command = "";
            string data = "";
            int space = 0;
            propertyValue = "";

            // The response string will be a command followed by a space followed by some data.  Parse these out.
            space = responseString.IndexOf(" ");

            // if the response contains no space then the response *is* the command; otherwise command is everything up to the space
            if (space == -1)
            {
                command = responseString;
            }
            else
            {
                command = responseString.Substring(0, space);
            }

            data = responseString.Substring(space + 1);

            switch (command)
            {
                case TheaterTekMessages.IP_FLASH:
                    //labelOSD.Text = data;
                    break;

                case TheaterTekMessages.IP_MEDIASTATE:
                    switch (data)
                    {
                        case "0":
                            propertyValue = "Stopped";
                            break;

                        case "1":
                            propertyValue = "Playing";
                            break;

                        case "2":
                            propertyValue = "Paused";
                            break;

                        case "3":
                            propertyValue = "Fast Forwarding";
                            break;

                        case "4":
                            propertyValue = "Fast Rewinding";
                            break;

                    }

                    Properties.GetPropertyByName("A/V Status").CurrentValue = propertyValue;

                    break;

                case TheaterTekMessages.IP_MEDIATIME:
                    Properties.GetPropertyByName("TheaterTek Time").CurrentValue = data.ToString();
                    //System.Diagnostics.Trace.WriteLine("Time is " + data.ToString());
                    break;

                case TheaterTekMessages.IP_CURFILE:
                    // Set the full path into the appropriate property
                    Properties.GetPropertyByName("TheaterTek Title Path").CurrentValue = data;

                    string movieTitle = "";

                    // Grab everything up to the first occurrence of \video_ts
                    Match titlePath = Regex.Match(data, @"(?<Front>.*)(\\video_ts)?", RegexOptions.IgnoreCase);

                    if (titlePath.Success)
                    {
                        // Reverse the string
                        movieTitle = ReverseString(titlePath.Groups["Front"].Value);

                        // Now find everything up *to* the first slash.  This will give me the title (but in reverse order).
                        Match titleOnly = Regex.Match(movieTitle, @"(?<Title>[^\\]*)", RegexOptions.IgnoreCase);

                        // Now reverse the string again for the final result.
                        if (titleOnly.Success) movieTitle = ReverseString(titleOnly.Groups["Title"].Value);

                        // If the title has a trailing ", The", move it to the beginning.
                        Match theString = Regex.Match(movieTitle, @"(?<Name>.*), The$", RegexOptions.IgnoreCase);
                        if (theString.Success) movieTitle = "The " + theString.Groups["Name"].Value;
                    }

                    //// grab everything after the last backslash (this *should* be the title)
                    //if (movieTitle.IndexOf(".") == -1)
                    //{
                    //    suffix = movieTitle.ToLower().LastIndexOf("\\");
                    //    movieTitle = movieTitle.Substring(suffix + 1);
                    //}
                    //else
                    //{
                    //    // in this case we must have an individual movie file.  For this case
                    //    // we want everything between the last backslash and the last period.
                    //    int slashPos;
                    //    int periodPos;

                    //    slashPos = movieTitle.LastIndexOf("\\");
                    //    periodPos = movieTitle.IndexOf(".");
                    //    if (slashPos != -1 & periodPos != -1)
                    //    {
                    //        movieTitle = movieTitle.Substring(slashPos + 1, periodPos - slashPos - 1);
                    //    }
                    //}

                    Properties.GetPropertyByName("TheaterTek Title").CurrentValue = movieTitle;
                    //System.Diagnostics.Trace.WriteLine("Title is " + movieTitle);
                    break;

                case TheaterTekMessages.IP_DVDTITLE:        // current "Title"
                    // When the title changes we'll ask TT for a chapter count for the new title
                    SendCommandToTT(TheaterTekMessages.IP_DVDCHAPTERCOUNT);
                    Properties.GetPropertyByName("TheaterTek Chapter").CurrentValue = data;
                    break;

                case TheaterTekMessages.IP_DVDCHAPTER:
                    Properties.GetPropertyByName("TheaterTek Chapter").CurrentValue = data;
                    break;

                case TheaterTekMessages.IP_DVDCHAPTERCOUNT:
                    Properties.GetPropertyByName("TheaterTek Chapter Count").CurrentValue = data;
                    break;

                default:
                    //System.Diagnostics.Trace.WriteLine("Command string is " + command);
                    //txtDataRx.Text += (responseString + "\r\n");
                    break;

            }
        }

        private void SendCommandToTT(string command)
        {
            if (command.Length == 5) command = "5000 " + command;
            command += "\n\r";      // the command MUST be terminated with a CRLF to be valid

            try
            {
                Object objData = command;
                byte[] byData = System.Text.Encoding.ASCII.GetBytes(objData.ToString());
                m_socClient.Send(byData);
            }
            catch (SocketException se)
            {
                System.Diagnostics.Trace.WriteLine("Error in SendCommandToTT: " + se.Message);
                //MessageBox.Show(se.Message);
            }
        }

        private void OpenSocket(string Address, int Port)
        {
            try
            {
                //create the socket instance...
                m_socClient = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);

                // get the remote IP address...
                IPAddress ip = Dns.GetHostEntry(Address).AddressList[0];
                int iPortNo = System.Convert.ToInt16(Port);

                //create the end point 
                IPEndPoint ipEnd = new IPEndPoint(ip, iPortNo);

                //connect to the remote host...
                m_socClient.Connect(ipEnd);

                Properties.GetPropertyByName("TheaterTek Running?").CurrentValue = "Yes";
                Properties.GetPropertyByName("TheaterTek PC Reachable?").CurrentValue = "Yes";

                //watch for data ( asynchronously )...
                WaitForData();
            }
            catch (SocketException se)
            {
                // host not reachable
                if (se.ErrorCode == 11001)
                {
                    Properties.GetPropertyByName("TheaterTek PC Reachable?").CurrentValue = "No";
                    Properties.GetPropertyByName("TheaterTek Running?").CurrentValue = "Unknown";
                    HBModule.TraceDeviceMessage(_deviceHandle, TRC_TYPE.ttError, "Unable to connect to TheaterTek host PC!");
                }

                // no response from the requested port (TT not running?)
                if (se.ErrorCode == 10061)
                {
                    // TODO: put try/catch block here to try starting TT using Autokiller
                    //       (will I run into recursive issues if I call OpenSocket again from here?)
                    Properties.GetPropertyByName("TheaterTek PC Reachable?").CurrentValue = "Yes";
                    Properties.GetPropertyByName("TheaterTek Running?").CurrentValue = "No";
                    HBModule.TraceDeviceMessage(_deviceHandle, TRC_TYPE.ttError, "Unable to connect to port on TheaterTek PC!");
                }

                System.Diagnostics.Trace.WriteLine("Error in OpenSocket: " + se.Message + "(" + se.ErrorCode + ")");
            }
        }

        private void WaitForData()
        {
            //System.Diagnostics.Trace.WriteLine("WaitForData triggered: ");
            try
            {
                if (pfnCallBack == null)
                {
                    pfnCallBack = new AsyncCallback(OnDataReceived);
                }
                CSocketPacket theSocPkt = new CSocketPacket();
                theSocPkt.thisSocket = m_socClient;
                // now start to listen for any data...
                m_asynResult = m_socClient.BeginReceive(theSocPkt.dataBuffer, 0, theSocPkt.dataBuffer.Length, SocketFlags.None, pfnCallBack, theSocPkt);
            }
            catch (SocketException se)
            {
                //MessageBox.Show(se.Message);
                System.Diagnostics.Trace.WriteLine("Error in WaitForData: " + se.Message);
            }
        }

        private class CSocketPacket
        {
            public System.Net.Sockets.Socket thisSocket;
            public byte[] dataBuffer = new byte[1];
        }

        private void OnDataReceived(IAsyncResult asyn)
        {
            try
            {
                //System.Diagnostics.Debugger.Log(0, "1", "\nOnDataReceived: triggered\n");
                Control.CheckForIllegalCrossThreadCalls = true;
                CSocketPacket theSockId = (CSocketPacket)asyn.AsyncState;
                //end receive...
                int iRx = 0;
                iRx = theSockId.thisSocket.EndReceive(asyn);
                char[] chars = new char[iRx + 1];
                System.Text.Decoder d = System.Text.Encoding.UTF8.GetDecoder();
                int charLen = d.GetChars(theSockId.dataBuffer, 0, iRx, chars, 0);
                System.String szData = new System.String(chars);
                //System.Diagnostics.Trace.WriteLine("szData (in OnDataReceived) = '" + szData + "'");

                if (szData.Substring(0, 1) == "\r")
                {
                    //System.Diagnostics.Trace.WriteLine("Response was " + responseString);
                    parseResponse(responseString);
                    responseString = "";
                }
                else
                {
                    if (szData.Substring(0, 1) != "\n" & szData.Substring(0, 1) != "\r") responseString += szData.Substring(0, 1);
                }

                WaitForData();
            }
            catch (ObjectDisposedException)
            {
                System.Diagnostics.Debugger.Log(0, "1", "\nOnDataReceived: Socket has been closed\n");
            }
            catch (SocketException se)
            {
                System.Diagnostics.Trace.WriteLine("Error in OnDataReceived: " + se.Message);
                //MessageBox.Show(se.Message);
            }
        }

        private void CloseSocket()
        {
            if (m_socClient != null && m_socClient.Connected)
            {
                m_socClient.Shutdown(SocketShutdown.Send);
                m_socClient.Close();
                m_socClient = null;
            }
        }

        private bool StartTheaterTek()
        {
            CloseSocket();

            try
            {
                OpenSocket(ttAddress, ttAKPort);
            }
            catch (Exception)
            {
                HBModule.TraceDeviceMessage(_deviceHandle, TRC_TYPE.ttError, "Error opening Autokiller socket to TheaterTek host PC!");
                return false;
            }

            SendCommandToTT(TheaterTekMessages.IP_LAUNCH);
            InitConnection();

            return true;
        }

        public static string ReverseString(string toReverse)
        {
            //
            // Convert the parameter string to a char array, then call Array's static
            // Reverse method, then return a new string based on the char[].
            //
            char[] charArray = toReverse.ToCharArray();
            Array.Reverse(charArray);
            return new string(charArray);
        }
    }
}
